! ===========================================================================
MODULE WMAP_PASS2

! This code is the central likelihood routine from which the other subroutines
! are called.
!
! Parameters are defined in associated WMAP_options module
!
! This code was a collaborative effort by the following WMAP Team Members:
! R. Bean
! E. Komatsu
! M. Nolta
! H. Peiris
! L. Verde
! D. Spergel
!
! ===========================================================================

  USE WMAP_options
  use tt_beam_and_ptsrc_chisq
  use WMAP_eebb_lowl
  use tlike

  character(len=*), parameter, public :: wmap_likelihood_version='v2p2'

  INTEGER, PARAMETER :: tt_lmax =1000 
  INTEGER, PARAMETER :: te_lmax = 450
  logical :: initialise_pass2=.true.

  REAL(8) :: R_off_tete(2:te_lmax, 2:te_lmax)
  REAL(8) :: R_off_tttt(2:tt_lmax, 2:tt_lmax)
  REAL(8) :: off_log_curv(2:tt_lmax, 2:tt_lmax)
  REAL(8) :: epsilon(2:tt_lmax,2:tt_lmax)

  REAL(8) :: cltt_dat(2:tt_lmax)
  REAL(8) :: clte_dat(2:te_lmax) 
  REAL(8) :: ntt(2:tt_lmax)    
  REAL(8) :: ntt_te(2:te_lmax)  
  REAL(8) :: nee_te(2:te_lmax)
  REAL(8) :: fskytt(2:tt_lmax)
  REAL(8) :: fskyte(2:te_lmax)

PRIVATE

PUBLIC PASS2_WMAP_init, PASS2_COMPUTE_LIKELIHOOD, pass2_wmap_degrees_of_freedom

contains

! ===========================================================================
SUBROUTINE PASS2_WMAP_init 
! ===========================================================================

  use wmap_util

  IMPLICIT NONE

  INTEGER  :: l,ll
  integer  :: il,ill,i,j
  REAL(8)  :: dummy
  CHARACTER(LEN=240) :: ttfilename,tefilename,ttofffilename,teofffilename
  LOGICAL  :: good
  integer ::  lun

#ifdef TIMING
	call wmap_timing_start( 'pass2_wmap_init' )
#endif

  print *, 'Initializing WMAP likelihood, version '//wmap_likelihood_version

!-----------------------------------------------
! initialise beam uncertainties
!-----------------------------------------------

  if ( use_TT_beam_ptsrc ) call init_tt_beam_and_ptsrc_chisq( 2, tt_lmax )

!-----------------------------------------------
! initialise low l codes
!-----------------------------------------------
  if(use_lowl_TT)call setup_for_TT_exact(lowl_max)

  if(use_lowl_pol)call eebb_lowl_like_setup
 
!-----------------------------------------------
! set file names
!-----------------------------------------------

  ttfilename = trim(WMAP_data_dir)//'wmap_likelihood_inputs_tt.dr2p2.dat'
  tefilename = trim(WMAP_data_dir)//'wmap_likelihood_inputs_te.p2.dat'
  ttofffilename = trim(WMAP_data_dir)//'wmap_likelihood_inputs_tt_offdiag.p2.dat'
  teofffilename = trim(WMAP_data_dir)//'wmap_likelihood_inputs_te_offdiag.p2.dat'

!-----------------------------------------------
! get TT diag
!-----------------------------------------------

  inquire(file=ttfilename,exist = good)
  if(.not.good)then
     write(*,*) 'cant find', trim(ttfilename)
     stop
  endif
  call get_free_lun( lun )
  open(unit=lun,file=ttfilename,form='formatted',status='unknown')

  do l=2,tt_lmax
     read(lun,*) dummy,cltt_dat(l),ntt(l),fskytt(l)
  enddo
  close(lun)

 
!-----------------------------------------------
! get TE diag
!-----------------------------------------------

  inquire(file=tefilename,exist = good)
  if(.not.good)then
     write(*,*) 'cant find', trim(tefilename)
     stop
  endif

  call get_free_lun( lun )
  open(unit=lun,file=tefilename,form='formatted',status='unknown')

  do l=2,te_lmax
     read(lun,*) ll,clte_dat(l), dummy, ntt_te(l),nee_te(l),fskyte(l)
  enddo
  close(lun)

!-----------------------------------------------
! get TT off diag
!-----------------------------------------------

  R_off_tttt = 0.

  inquire(file=ttofffilename,exist = good)

  if(.not.good)then
     write(*,*) 'cant find', trim(ttofffilename)
     stop
  endif

  call get_free_lun( lun )
  open(unit=lun,file=ttofffilename,form='formatted',status='unknown')

  do i=2,tt_lmax
     do j=i+1,tt_lmax

        read(lun,*) l,ll,epsilon(i,j),R_off_tttt(i,j)
        R_off_tttt(j,i)=R_off_tttt(i,j)
        epsilon(j,i)=epsilon(i,j)

        if(i.ne.l.or.j.ne.ll)then
           write(*,*)"tt off file misread",i,j,l,ll
           stop
        endif

     enddo
  enddo

  close(lun)

!-----------------------------------------------
! get TE off diag
!----------------------------------------------

  inquire(file=teofffilename,exist = good)

  if(.not.good)then
     write(*,*) 'cant find', trim(teofffilename)
     stop
  endif

  call get_free_lun( lun )
  open(unit=lun,file=teofffilename,form='formatted',status='unknown')

  do i=2,te_lmax-1
     do j=i+1,te_lmax

        read(lun,*) l ,ll,dummy

        if(l.le.te_lmax.and.ll.le.te_lmax)then
           R_off_tete(i,j)=dummy
           R_off_tete(j,i)=R_off_tete(i,j)
        endif

        if(l.ne.i.or.ll.ne.j)then
           write(*,*)"TE off diag misread i,j,l,ll",i,j,l,ll
           stop
        endif

     enddo
  enddo
  close(lun)

  initialise_pass2 = .false.

#ifdef TIMING
	call wmap_timing_end()
#endif

END SUBROUTINE PASS2_WMAP_init 

  subroutine pass2_wmap_degrees_of_freedom( tt_npix, teeebb_npix )

	integer, intent(out) :: tt_npix, teeebb_npix

	tt_npix = tt_pixlike_dof()
	teeebb_npix = teeebb_pixlike_dof()

  end subroutine

! ===========================================================================
SUBROUTINE PASS2_COMPUTE_LIKELIHOOD(cltt,clte,clee,clbb,like)
! ===========================================================================

  use wmap_util

  IMPLICIT NONE
  REAL(8), intent(in) :: cltt(2:*), clte(2:*), clee(2:*), clbb(2:*)
  REAL(8),intent(out) :: like(num_WMAP)
  INTEGER :: il, ill
  REAL(8) :: dlnlike_tot, dlnlike ,dlnlike_beam, ln_det_TETE
  !REAL(8) :: fisher(2:tt_lmax, 2:tt_lmax)
  REAL(8), allocatable :: fisher(:,:)
  REAL(8) :: tttt(2:tt_lmax),tete(2:te_lmax),ttte(2:te_lmax)
  REAL(8) :: z(2:tt_lmax), zbar(2:tt_lmax)
  REAL(8) :: cltt_temp(2:tt_lmax)
  real, allocatable,dimension(:) :: lowl_cl
  integer :: tt_hi_l_start, te_hi_l_start
  REAL(8) :: correlation_coefficient_cl

#ifdef TIMING
	call wmap_timing_start( 'pass2_compute_likelihood' )
#endif

  call wmap_likelihood_error_init

  if(initialise_pass2)then
     call PASS2_WMAP_init
  endif

  Like = 0.d0

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! allocate memory

  allocate( fisher(2:tt_lmax,2:tt_lmax) )

  !--------------------------------------------------------
  ! Are cltt, clte, and clee consistent? (EK, May 31, 2006)
  !--------------------------------------------------------
  do il=2,ttmax
     correlation_coefficient_cl = abs(clte(il))/sqrt(cltt(il)*clee(il))
     if(correlation_coefficient_cl>1)then
        print*,'unphysical input: TE/sqrt(TT*EE) > 1 at l=',il
        stop
     endif
  enddo

  cltt_temp(2:ttmax)=cltt(2:ttmax)

  !---------------------------------------------------------------------------
  ! low l TT likelihood
  !---------------------------------------------------------------------------
  if(use_lowl_TT)then
     allocate(lowl_cl(2:lowl_max))
     lowl_cl(2:lowl_max) = cltt_temp(2:lowl_max)
     call compute_tt_pixlike(lowl_max,lowl_cl,Like(ttlowllike),Like(ttlowldet))
     deallocate(lowl_cl)
     tt_hi_l_start = lowl_max+1
  else
     tt_hi_l_start = ttmin
  endif

  !---------------------------------------------------------------------------
  ! low l TE/EE/BB likelihood
  !---------------------------------------------------------------------------
  if(use_lowl_pol)then
     call eebb_lowl_likelihood(23,cltt_temp,clte,clee,clbb,Like(lowllike),Like(lowldet))
     te_hi_l_start = 24
  else
     te_hi_l_start = temin
  endif

  !---------------------------------------------------------------------------
  ! TT and TE covariance terms
  !---------------------------------------------------------------------------
  tttt=0.d0
  tete=0.d0
  ttte=0.d0

  if(use_TT.or.use_TE)then

     do il=ttmin,ttmax
        tttt(il)= 2.*(cltt_temp(il)+ntt(il))**2/((2.0d0*dble(il)+1.0d0)*fskytt(il)**2.)
     enddo

     do il=temin,temax
        tete(il)= ((cltt_temp(il)+ntt_te(il))*(clee(il)+nee_te(il))+clte(il)**2)/&
             ((2.0d0*dble(il)+1.0d0)*fskyte(il)**2.)
        ttte(il)= 2*((cltt_temp(il)+ntt(il))*clte(il))/&
             ((2.0d0*dble(il)+1.0d0)*fskyte(il)*fskytt(il))
     enddo

  endif

  !---------------------------------------------------------------------------
  !TTTT MASTER likelihood
  !---------------------------------------------------------------------------
  if(use_TT)then

     fisher=0.d0
     off_log_curv=0.d0

     do il=ttmin,ttmax
        z(il)=dlog(cltt_dat(il)+ntt(il))
        zbar(il)=dlog(cltt_temp(il)+ntt(il))
     enddo

     do il=ttmin,ttmax
        do ill=il,ttmax
           if(il.eq.ill)then
              if(il.le.te_lmax)then
                 fisher(il,ill) = tete(il)/(tttt(il)*tete(il)-ttte(il)**2)
              else
                 fisher(il,ill) = 1.d0/tttt(il)
              endif
           else
              fisher(il,ill) = -R_off_tttt(il,ill)/sqrt(tttt(il)*tttt(ill))&
                   +epsilon(il,ill)/(tttt(il)*tttt(ill))
           endif
           off_log_curv(il,ill)=(cltt_temp(il)+ntt(il))*fisher(il, ill)*&
                & (cltt_temp(ill)+ntt(ill))
        end do
     end do

     dlnlike_tot = 0.d0

     do il=ttmin,ttmax
        do ill=il,ttmax
  
           dlnlike = 2.d0/3.d0*(z(il)-zbar(il))*off_log_curv(il,ill)*(z(ill)-zbar(ill))+&
                1.d0/3.d0*(cltt_temp(il)-cltt_dat(il))* &
                fisher(il,ill)*(cltt_temp(ill)-cltt_dat(ill))

           if(il.ge.tt_hi_l_start.or.ill.ge.tt_hi_l_start)then
              if(il.eq.ill)then
                 dlnlike_tot = dlnlike_tot + dlnlike
              else
                 dlnlike_tot = dlnlike_tot + dlnlike*2
              endif
           endif

        end do
     end do
     
     Like(ttlike) = dlnlike_tot/2.d0

  endif
  !---------------------------------------------------------------------------
  !TTTT Beam and point source correction
  !---------------------------------------------------------------------------
  if(use_TT .and. use_TT_beam_ptsrc)then
	dlnlike_beam = compute_tt_beam_and_ptsrc_chisq( ttmin, ttmax, &
		cltt_temp, cltt_dat, ntt, fisher, z, zbar )
     if ( abs(dlnlike_beam) >= dlnlike_tot/4d0 ) then
        call wmap_likelihood_warning( 'beam correction invalid', 0 )
	dlnlike_beam = 0d0
     end if
     like(beamlike) = dlnlike_beam/2.d0
  endif

  !---------------------------------------------------------------------------
  !TETE MASTER likelihood
  !---------------------------------------------------------------------------
  if(use_TE)then
     ln_det_TETE=0.0d0
     dlnlike_tot=0.d0
     fisher=0.d0     

     do il=temin,temax
        do ill=il,temax
           if(il.eq.ill)then
              if(il.le.te_lmax)then
                 fisher(il,ill) = tttt(il)/(tttt(il)*tete(il)-ttte(il)**2)
              else
                 fisher(il,ill) = 1.d0/tete(il)
              endif
           else
              fisher(il,ill) = -R_off_tete(il,ill)/sqrt(tete(il)*tete(ill))
           endif
        enddo
     enddo

     do il=temin,temax
        do ill=il,temax

           dlnlike = (clte(il)-clte_dat(il))*fisher(il,ill)* (clte(ill)-clte_dat(ill))
           
           if(il.ge.te_hi_l_start.or.ill.ge.te_hi_l_start)then
              IF(il.eq.ill) then
                 dlnlike_tot=dlnlike_tot+dlnlike
                 ln_det_TETE=ln_det_TETE-Log(fisher(il,ill))
              else
                 dlnlike_tot=dlnlike_tot+dlnlike*2.
              endif
           endif

        end do
     end do
     Like(telike) = dlnlike_tot/2.d0
     
     Like(tedet) = (ln_det_TETE - te_lndet_offset)/2d0
  endif

10 continue

  ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! deallocate memory

  deallocate( fisher )

#ifdef TIMING
	call wmap_timing_end()
#endif

end SUBROUTINE PASS2_COMPUTE_LIKELIHOOD

END MODULE WMAP_PASS2

